# pylint: disable=R0913
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.enums as enums


class PlotSettings:
    def __init__(
        self,
        chart=enums.PlotCharts.MAIN_CHART.value,
        x_multiplier=1000,
        kind="scattergl",
        mode="markers",
        y_data=None,
    ):
        self.chart = chart
        self.x_multiplier = x_multiplier
        self.kind = kind
        self.mode = mode
        self.y_data = y_data



# pylint: disable=C0103,W0703,C0415
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import abc
import copy


import octobot_commons.enums as commons_enums
import octobot_commons.configuration as configuration
import octobot_commons.databases as databases


class AbstractTentacle:
    """
    The parent class of any OctoBot tentacle
    """

    __metaclass__ = abc.ABCMeta
    ALLOW_SUPER_CLASS_CONFIG = False
    USER_INPUT_TENTACLE_TYPE = commons_enums.UserInputTentacleTypes.UNDEFINED

    def __init__(self):
        self.logger = None
        self.UI: configuration.UserInputFactory = configuration.UserInputFactory(
            self.USER_INPUT_TENTACLE_TYPE
        )
        self.UI.set_tentacle_class(self.__class__).set_tentacle_config_proxy(
            self.get_local_config
        )

    @classmethod
    def get_name(cls) -> str:
        """
        Tentacle name based on class name
        :return: the tentacle name
        """
        return cls.__name__

    @classmethod
    def get_all_subclasses(cls) -> list:
        """
        Return all subclasses of this tentacle
        :return: the subclasses
        """
        subclasses_list = cls.__subclasses__()
        if cls.__subclasses__():
            for subclass in copy.deepcopy(subclasses_list):
                subclasses_list += subclass.get_all_subclasses()
        return subclasses_list

    @classmethod
    def get_user_commands(cls) -> dict:
        """
        Return the dict of user commands for this tentacle
        :return: the commands dict
        """
        return {}

    def get_local_config(self):
        """
        :return: the config of the tentacle
        """
        raise NotImplementedError

    @classmethod
    def create_local_instance(cls, config, tentacles_setup_config, tentacle_config):
        """
        :param config: the global configuration to give to the tentacle
        :param tentacles_setup_config: the global tentacles setup configuration to give to the tentacle
        :param tentacle_config: the tentacle configuration to give to the tentacle
        :return: a local, aimed to be short-lived, tentacle instance
        """
        raise NotImplementedError

    def init_user_inputs(self, inputs: dict) -> None:
        """
        Called right before starting the tentacle, should define all the tentacle's user inputs unless
        those are defined somewhere else.
        """

    async def load_and_save_user_inputs(self, bot_id: str):
        """
        Initialize and save the user inputs of the tentacle
        """
        try:
            inputs = {}
            self.init_user_inputs(inputs)
            if databases.RunDatabasesProvider.instance().is_storage_enabled(bot_id):
                run_db = databases.RunDatabasesProvider.instance().get_run_db(bot_id)
                await configuration.clear_user_inputs(run_db, self.get_name())
                for user_input in inputs.values():
                    await configuration.save_user_input(user_input, run_db)
                await run_db.flush()
        except Exception as err:
            self.logger.exception(
                err, True, f"Error when initializing user inputs: {err}"
            )

    @classmethod
    async def get_raw_config_and_user_inputs(
        cls, config, tentacles_setup_config, bot_id
    ):
        """
        :return: the tentacle configuration and its list of user inputs
        """
        try:
            import octobot_tentacles_manager.api as api

            specific_config = api.get_tentacle_config(tentacles_setup_config, cls)
        except ImportError as err:
            raise ImportError("octobot_tentacles_manager is required") from err
        if saved_user_inputs := await configuration.get_user_inputs(
            databases.RunDatabasesProvider.instance().get_run_db(bot_id),
            cls.get_name(),
        ):
            # user inputs have been saved in run database, use those as they might contain additional
            # (nested) user inputs
            return specific_config, saved_user_inputs
        # use user inputs from init_user_inputs
        tentacle_instance = cls.create_local_instance(
            config, tentacles_setup_config, specific_config
        )
        user_inputs = {}
        tentacle_instance.init_user_inputs(user_inputs)
        return specific_config, list(
            user_input.to_dict() for user_input in user_inputs.values()
        )



# pylint: disable=R0913, R0902, W0703
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import functools
import copy
import shutil
import octobot_commons.logging as logging
import octobot_commons.errors as errors
import octobot_commons.constants as commons_constants
import octobot_commons.profiles as profiles
import octobot_commons.json_util as json_util
import octobot_commons.configuration.config_file_manager as config_file_manager
import octobot_commons.configuration.config_operations as config_operations


class Configuration:
    """
    Configuration is managing an OctoBot configuration regarding reading, writing and updating
    """

    def __init__(
        self,
        config_path: str,
        profiles_path: str,
        schema_path: str = None,
        profile_schema_path: str = None,
    ):
        self.logger = logging.get_logger(self.__class__.__name__)
        self.config_path: str = config_path
        self.profiles_path: str = profiles_path
        self.config: dict = None
        self.config_schema_path: str = (
            schema_path or commons_constants.CONFIG_FILE_SCHEMA
        )
        self.profile_schema_path: str = (
            profile_schema_path or commons_constants.PROFILE_FILE_SCHEMA
        )

        self._read_config: dict = None
        self.profile: profiles.Profile = None
        self.profile_by_id: dict = {}

    def validate(self) -> None:
        """
        Validated self._read_config and self._profile against their json schema
        :return: None
        """
        json_util.validate(self._read_config, self.config_schema_path)
        self.profile.validate()

    def read(self, should_raise=True, fill_missing_fields=False) -> None:
        """
        Reads the configuration from self.config_path and load the current profile
        Overall config is stored into self.config and consists of a merger from the user
        config and activated profile
        :param should_raise: will raise upon exception when True
        :param fill_missing_fields: will try to fill in missing fields when true
        :return: None
        """
        self._read_config = config_file_manager.load(
            self.config_path,
            should_raise=should_raise,
            fill_missing_fields=fill_missing_fields,
        )
        self.config = copy.deepcopy(self._read_config)
        self.load_profiles_if_possible_and_necessary()

    def load_profiles_if_possible_and_necessary(self) -> None:
        """
        Loads profiles if profiles already exists and have not been already loaded
        :return: None
        """
        if not self.are_profiles_empty_or_missing() and not self.are_profile_loaded():
            self.load_profiles()
            self.select_profile(self._get_selected_profile())

    def select_profile(self, profile_id) -> None:
        """
        Sets self.profile using its profile_id
        :param profile_id: id of the profile to select
        :return: None
        """
        self.config[commons_constants.CONFIG_PROFILE] = profile_id
        self.profile = self.profile_by_id[profile_id]
        self.logger.info(f"Using {self.profile.name} profile.")
        self._generate_config_from_user_config_and_profile()

    def remove_profile(self, profile_id: str) -> None:
        """
        Removes the given profile and deletes its folder on disk
        :param profile_id: the id of the profile to remove
        :return: None
        """
        profile = self.profile_by_id[profile_id]
        if profile.read_only:
            raise errors.ProfileRemovalError(f"{profile.name} profile can't be removed")
        try:
            shutil.rmtree(profile.path)
            self.profile_by_id.pop(profile_id, None)
        except Exception as err:
            raise errors.ProfileRemovalError() from err

    def _generate_config_from_user_config_and_profile(self):
        for profile_managed_element in self.profile.FULLY_MANAGED_ELEMENTS:
            self.config[profile_managed_element] = copy.deepcopy(
                self.profile.config[profile_managed_element]
            )
        for partially_managed_element in self.profile.PARTIALLY_MANAGED_ELEMENTS:
            self.profile.merge_partially_managed_element_into_config(
                self.config, partially_managed_element
            )

    def save(
        self,
        temp_restore_config_file=commons_constants.TEMP_RESTORE_CONFIG_FILE,
        schema_file=None,
        sync_all_profiles=False,
    ) -> None:
        """
        Save the current self.config and self.profile.
        Synchronize all profiles if sync_all_profiles
        :return: None
        """
        config_to_save = self._get_config_without_profile_elements()
        config_file_manager.dump(
            self.config_path,
            config_to_save,
            temp_restore_config_file=temp_restore_config_file,
            schema_file=schema_file,
        )
        if self.profile is not None:
            self.profile.save_config(self.config)
        if sync_all_profiles:
            self._sync_other_profiles()

    def _sync_other_profiles(self):
        """
        Update profile partially managed elements for all profiles except self.profile
        with self.config
        """
        for profile in self.profile_by_id.values():
            if profile is self.profile:
                # do not synchronize self.profile
                continue
            profile.remove_deleted_elements(self.config)
            profile.validate_and_save_config()

    def is_loaded(self) -> bool:
        """
        Checks if self has been loaded
        :return: True when self has been loaded (read)
        """
        return self.config is not None

    def is_config_file_empty_or_missing(self) -> bool:
        """
        Checks if self.config_path existing and not empty
        :return: True when self.config_path is existing and non empty
        """
        return (not os.path.isfile(self.config_path)) or os.stat(
            self.config_path
        ).st_size == 0

    def are_profile_loaded(self) -> bool:
        """
        Checks if profiles have already been loaded
        :return: True if profiles have been loaded
        """
        return self.profile is not None

    def are_profiles_empty_or_missing(self) -> bool:
        """
        Checks if self.profiles_path exists and contains folders
        :return: True if profiles folder is not empty
        """
        return not (
            os.path.isdir(self.profiles_path) and os.listdir(self.profiles_path)
        )

    def get_non_imported_profiles(self) -> list:
        """
        :return: The list of loaded profiles in self that have not been imported into this OctoBot
        """
        return [
            profile for profile in self.profile_by_id.values() if not profile.imported
        ]

    def get_tentacles_config_path(self) -> str:
        """
        :return: The tentacles configurations associated to the activated profile
        """
        return self.profile.get_tentacles_config_path()

    def get_metrics_enabled(self) -> bool:
        """
        Check if metrics are enabled
        :return: True if metrics are enabled
        """
        return bool(
            self.config.get(commons_constants.CONFIG_METRICS, {}).get(
                commons_constants.CONFIG_ENABLED_OPTION, True
            )
        )

    def get_metrics_id(self) -> str:
        """
        :return: The current user's metrics id
        """
        return self.config[commons_constants.CONFIG_METRICS][
            commons_constants.CONFIG_METRICS_BOT_ID
        ]

    def accepted_terms(self) -> bool:
        """
        Check if terms has been accepted
        :return: the check result
        """
        return self.config.get(commons_constants.CONFIG_ACCEPTED_TERMS, False)

    def accept_terms(self, accepted) -> None:
        """
        Perform terms acceptation
        :param accepted: accepted or not
        """
        self.config[commons_constants.CONFIG_ACCEPTED_TERMS] = accepted
        self.save()

    def update_config_fields(
        self,
        to_update_fields,
        in_backtesting,
        config_separator,
        delete=False,
    ) -> None:
        """
        Partially update self.config using the fields found in to_update_fields
        :param to_update_fields: the fields to update
        :param in_backtesting: if backtesting is enabled
        :param config_separator: the config separator
        :param delete: if the data should be removed
        """
        config_operations.filter_to_update_data(to_update_fields, in_backtesting)
        removed_configs = []
        if delete:
            removed_configs = [
                config_operations.parse_and_update(
                    data_key, config_operations.DELETE_ELEMENT_VALUE, config_separator
                )
                for data_key in to_update_fields
            ]
            functools.reduce(
                config_operations.clear_dictionaries_by_keys,
                [self.config] + removed_configs,
            )
        else:
            updated_configs = [
                config_operations.parse_and_update(
                    data_key, data_value, config_separator
                )
                for data_key, data_value in to_update_fields.items()
            ]
            # merge configs
            functools.reduce(
                config_operations.merge_dictionaries_by_appending_keys,
                [self.config] + updated_configs,
            )

        # save config
        self.save(
            schema_file=self.config_schema_path, sync_all_profiles=bool(removed_configs)
        )

    def _get_selected_profile(self):
        selected_profile_id = self._read_config.get(
            commons_constants.CONFIG_PROFILE, commons_constants.DEFAULT_PROFILE
        )
        if selected_profile_id in self.profile_by_id:
            return selected_profile_id
        if (
            selected_profile_id != commons_constants.DEFAULT_PROFILE
            and commons_constants.DEFAULT_PROFILE in self.profile_by_id
        ):
            return commons_constants.DEFAULT_PROFILE
        raise errors.NoProfileError

    def load_profiles(self) -> None:
        """
        Loads the available profiles
        :return: None
        """
        for profile in profiles.Profile.get_all_profiles(
            self.profiles_path, schema_path=self.profile_schema_path
        ):
            if profile.profile_id not in self.profile_by_id:
                self.profile_by_id[profile.profile_id] = profile

    def _get_config_without_profile_elements(self) -> dict:
        filtered_config = copy.deepcopy(self.config)
        # do not include profile fully managed elements into filtered config
        for profile_managed_element in profiles.Profile.FULLY_MANAGED_ELEMENTS:
            filtered_config.pop(profile_managed_element, None)
        return filtered_config

    def dev_mode_enabled(self) -> bool:
        """
        Check if DEV_MODE is enabled
        :return: bool
        """
        return os.getenv(
            commons_constants.CONFIG_DEBUG_OPTION,
            self.config.get(commons_constants.CONFIG_DEBUG_OPTION, False),
        )



# cython: language_level=3
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


cpdef object parse_symbol(str symbol)   # return object to allow exception propagation
cpdef str merge_symbol(str symbol)
cpdef str merge_currencies(
        str currency,
        str market,
        str settlement_asset=*,
        str market_separator=*,
        str settlement_separator=*
)
cpdef str convert_symbol(
    str symbol,
    str symbol_separator,
    str new_symbol_separator=*,
    bint should_uppercase=*,
    bint should_lowercase=*,
    bint base_and_quote_only=*
)



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.singleton as singleton
import octobot_commons.databases.implementations.meta_database as meta_database
import octobot_commons.errors as errors
import octobot_commons.logging as logging


class RunDatabasesProvider(singleton.Singleton):
    def __init__(self):
        self.logger = logging.get_logger(self.__class__.__name__)
        self.run_databases = {}

    async def add_bot_id(
        self, bot_id, run_database_identifier, with_lock=False, cache_size=None
    ):
        """
        Initialize the given run_database_identifier and create a new MetaDatabase associated to the given bot_id
        """
        await run_database_identifier.initialize()
        self.run_databases[bot_id] = meta_database.MetaDatabase(
            run_database_identifier, with_lock=with_lock, cache_size=cache_size
        )

    def has_bot_id(self, bot_id):
        """
        :return: True if the given bot_id has been added via add_bot_id
        """
        return bot_id in self.run_databases

    def is_storage_enabled(self, bot_id):
        """
        :return: True if storage is enabled for the given bot_id
        """
        return self.run_databases[bot_id].run_dbs_identifier.enable_storage

    def get_any_run_databases_identifier(self):
        """
        :return: the first added run_dbs_identifier
        """
        return next(iter(self.run_databases.values())).run_dbs_identifier

    def get_run_databases_identifier(self, bot_id):
        """
        :return: the bot_id associated run_dbs_identifier
        """
        return self.run_databases[bot_id].run_dbs_identifier

    def get_run_db(self, bot_id):
        """
        :return: the bot_id associated run database
        """
        return self.run_databases[bot_id].get_run_db()

    def get_orders_db(self, bot_id, account_type, exchange=None):
        """
        :return: the bot_id and exchange associated orders database and account_type.
        Use local run_database_identifier.context for exchange if not provided.
        """
        return self.run_databases[bot_id].get_orders_db(account_type, exchange=exchange)

    def get_trades_db(self, bot_id, account_type, exchange=None):
        """
        :return: the bot_id and exchange associated trades database and account_type.
        Use local run_database_identifier.context for exchange if not provided.
        """
        return self.run_databases[bot_id].get_trades_db(account_type, exchange=exchange)

    def get_transactions_db(self, bot_id, account_type, exchange=None):
        """
        :return: the bot_id and exchange associated transactions database and account_type.
        Use local run_database_identifier.context for exchange if not provided.
        """
        return self.run_databases[bot_id].get_transactions_db(
            account_type, exchange=exchange
        )

    def get_backtesting_metadata_db(self, bot_id):
        """
        :return: the bot_id and exchange associated backtesting metadata database.
        """
        return self.run_databases[bot_id].get_backtesting_metadata_db()

    def get_symbol_db(self, bot_id, exchange, symbol):
        """
        :return: the bot_id and exchange associated transactions database.
        Use local run_database_identifier.context for exchange if exchange is None.
        """
        if not symbol:
            raise errors.DatabaseNotFoundError("symbol parameter has to be provided")
        return self.run_databases[bot_id].get_symbol_db(exchange, symbol)

    async def get_all_symbol_dbs(self, bot_id, exchange):
        """
        :return: an iterable over each symbol database for the given exchange
        """
        return await self.run_databases[bot_id].get_all_symbol_dbs(exchange)

    def get_historical_portfolio_value_db(self, bot_id, account_type, exchange):
        """
        :return: the bot_id, exchange and account_type associated transactions database.
        """
        return self.run_databases[bot_id].get_historical_portfolio_value_db(
            account_type, exchange
        )

    async def close(self, bot_id):
        """
        Close the bot_id associated databases. Does not pop bot_id from self.run_databases to allow post-close calls.
        """
        self.logger.debug(f"Closing bot storage for bot_id: {bot_id} ...")
        await self.run_databases[bot_id].close()
        # do not pop bot_id to keep run data access
        self.logger.debug(f"Closed bot storage for bot_id: {bot_id}")



# pylint: disable=C0116,R0801
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy
import sortedcontainers

import octobot_commons.enums as enums
import octobot_commons.databases.implementations.db_writer as writer
import octobot_commons.databases.document_database_adaptors as adaptors


class CacheDatabase(writer.DBWriter):
    CACHE_TABLE = enums.CacheDatabaseTables.CACHE.value
    CACHE_METADATA_TABLE = enums.CacheDatabaseTables.METADATA.value
    UUID_KEY = "uuid"

    def __init__(
        self,
        file_path: str,
        database_adaptor=adaptors.TinyDBAdaptor,
        cache_size=None,
        **kwargs
    ):
        super().__init__(
            file_path,
            database_adaptor=database_adaptor,
            cache_size=cache_size,
            **kwargs
        )
        self._are_metadata_written = False
        self._local_cache = None
        self.metadata = {
            enums.CacheDatabaseColumns.TYPE.value: self.__class__.__name__,
        }

    def get_non_default_metadata(self):
        metadata = copy.copy(self.metadata)
        metadata.pop(enums.CacheDatabaseColumns.TYPE.value)
        return metadata

    def add_metadata(self, additional_metadata: dict):
        self.metadata.update(additional_metadata)

    async def _ensure_metadata(self):
        if not self._are_metadata_written:
            await self._database.upsert(
                self.CACHE_METADATA_TABLE, self.metadata, None, uuid=1
            )
            self._are_metadata_written = True

    async def _ensure_local_cache(self, identifier_key, update=False):
        if update or self._local_cache is None:
            self._local_cache = sortedcontainers.SortedDict()
            for cache in await self.get_cache():
                cache[self.UUID_KEY] = self._database.get_uuid(cache)
                self._local_cache[cache[identifier_key]] = cache

    async def get_metadata(self):
        return await self._database.select(self.CACHE_METADATA_TABLE, None, uuid=1)

    async def get_cache(self):
        return await self._database.select(self.CACHE_TABLE, None)

    async def clear(self):
        await self._database.delete(self.CACHE_TABLE, None)
        await self._database.delete(self.CACHE_METADATA_TABLE, None)
        await super().clear()
        self._local_cache = sortedcontainers.SortedDict()
        self._are_metadata_written = False
        # always rewrite metadata as they are necessary to handle cache later
        await self._ensure_metadata()
        await self.flush()

    async def _get_from_local_cache(self, identifier_key, identifier_value, sub_key):
        await self._ensure_local_cache(identifier_key)
        return self._local_cache[identifier_value][sub_key]

    async def _needs_update(
        self, identifier_key, identifier_value, sub_key, value
    ) -> bool:
        try:
            return (
                await self._get_from_local_cache(
                    identifier_key, identifier_value, sub_key
                )
                != value
            )
        except KeyError:
            return True



# pylint: disable=R0902,R0913,C0415,R0904
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import shutil

import octobot_commons.databases.document_database_adaptors as adaptors
import octobot_commons.constants as constants
import octobot_commons.enums as enums
import octobot_commons.symbols.symbol_util as symbol_util


class RunDatabasesIdentifier:
    def __init__(
        self,
        tentacle_class,
        optimization_campaign_name=None,
        database_adaptor=adaptors.TinyDBAdaptor,
        backtesting_id=None,
        live_id=None,
        optimizer_id=None,
        context=None,
        enable_storage=True,
    ):
        self.database_adaptor = database_adaptor
        self.optimization_campaign_name = optimization_campaign_name
        self.backtesting_id = backtesting_id
        self.live_id = live_id
        self.optimizer_id = optimizer_id
        self.tentacle_class = (
            tentacle_class
            if isinstance(tentacle_class, str)
            else tentacle_class.__name__
        )
        self.enable_storage = enable_storage
        self.context = context
        self.data_path = self._merge_parts(constants.USER_FOLDER, constants.DATA_FOLDER)
        self.base_path = self._merge_parts(self.data_path, self.tentacle_class)
        self.suffix = (
            self.database_adaptor.get_db_file_ext()
            if self.database_adaptor.is_file_system_based()
            else ""
        )

    async def initialize(self, exchange=None):
        """
        Initializes the necessary elements for these run databases. Creates necessary folder on file system databases
        :param exchange: name of the associated exchange
        Used for live trading cross trading mode stats (such as profitability)
        """
        if not self.enable_storage:
            return
        # global history is a live only feature
        from_global_history = self.backtesting_id is None
        deepest_identifier = (
            self._base_folder(from_global_history=from_global_history)
            if exchange is None
            else self._merge_parts(
                self._base_folder(from_global_history=from_global_history), exchange
            )
        )
        await self.database_adaptor.create_identifier(deepest_identifier)

    def get_run_data_db_identifier(self) -> str:
        """
        :return: the database identifier associated to the run database
        """
        return self._get_db_identifier(enums.RunDatabases.RUN_DATA_DB.value, None)

    def get_orders_db_identifier(self, account_type, exchange) -> str:
        """
        :return: the database identifier associated to this exchange's orders
        :param account_type: type of account
        :param exchange: name of the associated exchange
        """
        return self._get_db_identifier(
            f"{enums.RunDatabases.ORDERS_DB.value}{account_type}", exchange
        )

    def get_trades_db_identifier(self, account_type, exchange) -> str:
        """
        :return: the database identifier associated to this exchange's trades
        :param account_type: type of account
        :param exchange: name of the associated exchange
        """
        return self._get_db_identifier(
            f"{enums.RunDatabases.TRADES_DB.value}{account_type}", exchange
        )

    def get_transactions_db_identifier(self, account_type, exchange) -> str:
        """
        :return: the database identifier associated to this exchange's transactions
        :param account_type: type of account
        :param exchange: name of the associated exchange
        """
        return self._get_db_identifier(
            f"{enums.RunDatabases.TRANSACTIONS_DB.value}{account_type}", exchange
        )

    def get_symbol_db_identifier(self, exchange, symbol) -> str:
        """
        :return: the database identifier associated to this exchange's symbol data
        :param exchange: name of the associated exchange
        :param symbol: the associated symbol
        """
        return self._get_db_identifier(symbol_util.merge_symbol(symbol), exchange)

    def get_historical_portfolio_value_db_identifier(
        self, account_type, exchange
    ) -> str:
        """
        :return: the database identifier associated to this exchange's historical portfolio value
        :param account_type: a suffix identifying the type of portfolio (future / sandbox etc)
        :param exchange: name of the associated exchange
        """
        return self._get_db_identifier(
            f"{enums.RunDatabases.PORTFOLIO_VALUE_DB.value}{account_type}",
            exchange,
        )

    def get_backtesting_metadata_identifier(self) -> str:
        """
        :return: the database identifier associated to backtesting metadata
        """
        return self._get_db_identifier(
            enums.RunDatabases.METADATA.value, None, ignore_backtesting_id=True
        )

    def get_bot_live_metadata_identifier(self) -> str:
        """
        :return: the database identifier associated to live metadata
        """
        return self._get_db_identifier(
            enums.RunDatabases.METADATA.value, None, ignore_live_id=True
        )

    def _get_db_identifier(self, run_database_name, exchange, **base_folder_kwargs):
        if exchange is None:
            return self._merge_parts(
                self._base_folder(**base_folder_kwargs),
                self.get_db_full_name(run_database_name),
            )
        return self._merge_parts(
            self._base_folder(**base_folder_kwargs),
            exchange,
            self.get_db_full_name(run_database_name),
        )

    def get_db_full_name(self, db_name):
        """
        :return: the db_name's associated database name including suffix
        """
        return f"{db_name}{self.suffix}"

    async def exchange_base_identifier_exists(self, exchange) -> bool:
        """
        :return: True if there are data under this exchange name
        """
        return await self.database_adaptor.identifier_exists(
            self.get_exchange_based_identifier(exchange), False
        )

    def get_exchange_based_identifier(self, exchange):
        """
        :return: the database identifier associated to the given exchange
        """
        return self._merge_parts(self._base_folder(), exchange)

    async def get_single_existing_exchange(self) -> str:
        """
        :return: the name of the only exchange the backtesting happened on if it only ran on a single exchange,
        None otherwise
        """
        ignored_folders = [enums.RunDatabases.LIVE.value]
        try:
            import octobot_tentacles_manager.constants as tentacles_manager_constants

            ignored_folders.append(
                tentacles_manager_constants.TENTACLES_SPECIFIC_CONFIG_FOLDER
            )
        except ImportError:
            pass
        return await self.database_adaptor.get_single_sub_identifier(
            self._base_folder(), ignored_folders
        )

    async def symbol_base_identifier_exists(self, exchange, symbol) -> bool:
        """
        :return: True if there are data under this exchange name
        """
        identifier = self._merge_parts(
            self._base_folder(),
            exchange,
            self.get_db_full_name(symbol_util.merge_symbol(symbol)),
        )
        return await self.database_adaptor.identifier_exists(identifier, True)

    def get_backtesting_run_folder(self) -> str:
        """
        :return: base folder associated to a backtesting run
        """
        return self._base_folder()

    def get_optimizer_runs_schedule_identifier(self) -> str:
        """
        :return: the identifier associated to the optimizer run schedule database
        """
        return self._merge_parts(
            self.base_path,
            self.optimization_campaign_name,
            enums.RunDatabases.OPTIMIZER.value,
            self.get_db_full_name(enums.RunDatabases.OPTIMIZER_RUNS_SCHEDULE_DB.value),
        )

    async def generate_new_backtesting_id(self) -> int:
        """
        :return: a new unique backtesting id
        """
        return await self._generate_new_id(is_optimizer=False)

    async def generate_new_bot_live_id(self) -> int:
        """
        :return: a new unique bot recording id
        """
        return await self._generate_new_id(is_optimizer=False, is_bot_recording=True)

    async def generate_new_optimizer_id(self, back_list) -> int:
        """
        :return: a new unique optimizer id
        """
        return await self._generate_new_id(back_list=back_list, is_optimizer=True)

    def is_symbol_database(self, database_identifier: str) -> bool:
        """
        :return: True if the given identifier is related to a symbol database
        """
        return database_identifier.endswith(self.suffix) and all(
            not other_identifier.value in database_identifier
            for other_identifier in enums.RunDatabases
        )

    def get_symbol_db_name(self, symbol_db_identifier):
        """
        :return: the given identifier's database name (without suffix if any)
        """
        return symbol_db_identifier.split(self.suffix)[0]

    def remove_all(self):
        """
        Clears every data from a backtesting run
        """
        identifier = self._base_folder()
        if self.database_adaptor.is_file_system_based():
            if os.path.isdir(identifier):
                shutil.rmtree(identifier)
            return
        raise RuntimeError(f"Unhandled database_adaptor {self.database_adaptor}")

    async def _generate_new_id(
        self, back_list=None, is_optimizer=False, is_bot_recording=False
    ):
        back_list = back_list or []
        max_runs = (
            constants.MAX_OPTIMIZER_RUNS
            if is_optimizer
            else constants.MAX_BACKTESTING_RUNS
        )
        for index in range(1, max_runs + 1):
            if index in back_list:
                continue
            name_candidate = (
                self._base_folder(optimizer_id=index)
                if is_optimizer
                else (
                    self._base_folder(live_id=index)
                    if is_bot_recording
                    else self._base_folder(backtesting_id=index)
                )
            )
            if not await self.database_adaptor.identifier_exists(name_candidate, False):
                return index
        raise RuntimeError(
            f"Reached maximum number of {'optimizer' if is_optimizer else 'backtesting'} runs "
            f"({constants.MAX_BACKTESTING_RUNS}). Please remove some."
        )

    async def get_optimization_campaign_names(self) -> list:
        """
        :return: a list of every existing campaign name
        """
        optimization_campaign_folder = self._merge_parts(self.base_path)
        if await self.database_adaptor.identifier_exists(
            optimization_campaign_folder, False
        ):
            return [
                element
                async for element in self.database_adaptor.get_sub_identifiers(
                    optimization_campaign_folder, [enums.RunDatabases.LIVE.value]
                )
            ]
        return []

    async def get_optimizer_run_ids(self) -> list:
        """
        :return: a list of every optimizer id in the current campaign
        """
        optimizer_runs_path = self._merge_parts(
            self.base_path,
            self.optimization_campaign_name,
            enums.RunDatabases.OPTIMIZER.value,
        )
        if await self.database_adaptor.identifier_exists(optimizer_runs_path, False):
            return [
                self.parse_optimizer_id(element)
                async for element in self.database_adaptor.get_sub_identifiers(
                    optimizer_runs_path, []
                )
            ]

    async def get_backtesting_run_ids(self) -> list:
        """
        :return: a list of every backtesting id in the current campaign
        """
        runs_path = self._base_folder(ignore_backtesting_id=True)
        if await self.database_adaptor.identifier_exists(runs_path, False):
            return [
                self.parse_backtesting_id(element)
                async for element in self.database_adaptor.get_sub_identifiers(
                    runs_path, []
                )
            ]

    @staticmethod
    def parse_optimizer_id(identifier) -> str:
        """
        :return: the associated optimizer id
        """
        return identifier.split(constants.DB_SEPARATOR)[-1]

    @staticmethod
    def parse_backtesting_id(identifier) -> str:
        """
        :return: the associated backtesting id
        """
        return identifier.split(constants.DB_SEPARATOR)[-1]

    def _get_base_path(self, from_global_history, backtesting_id, optimizer_id):
        if from_global_history and (backtesting_id is None and optimizer_id is None):
            # in live global history, use self.data_path as it's not related to a trading mode
            return self.data_path
        return self.base_path

    def _base_folder(
        self,
        ignore_backtesting_id=False,
        backtesting_id=None,
        live_id=None,
        ignore_live_id=None,
        ignore_optimizer_id=False,
        optimizer_id=None,
        from_global_history=True,
    ) -> str:
        backtesting_id = backtesting_id or self.backtesting_id
        optimizer_id = optimizer_id or self.optimizer_id
        path = self._get_base_path(from_global_history, backtesting_id, optimizer_id)
        live_id = live_id or self.live_id
        # when in optimizer or backtesting: wrap it into the current campaign
        if backtesting_id is not None or optimizer_id is not None:
            if self.optimization_campaign_name is None:
                raise RuntimeError(
                    f"optimization_campaign_name is required in {RunDatabasesIdentifier} "
                    f"constructor while in a backtesting or optimizer context"
                )
            path = self._merge_parts(path, self.optimization_campaign_name)
        if optimizer_id is not None:
            if ignore_optimizer_id:
                path = self._merge_parts(path, enums.RunDatabases.OPTIMIZER.value)
            else:
                path = self._merge_parts(
                    path,
                    enums.RunDatabases.OPTIMIZER.value,
                    f"{enums.RunDatabases.OPTIMIZER.value}{constants.DB_SEPARATOR}{optimizer_id}",
                )
        if backtesting_id is not None:
            if optimizer_id is None:
                path = self._merge_parts(path, enums.RunDatabases.BACKTESTING.value)
            if ignore_backtesting_id:
                return path
            return self._merge_parts(
                path,
                f"{enums.RunDatabases.BACKTESTING.value}"
                f"{constants.DB_SEPARATOR}{backtesting_id}",
            )
        if optimizer_id is None:
            # live mode
            if ignore_live_id:
                return self._merge_parts(path, enums.RunDatabases.LIVE.value)
            return self._merge_parts(
                path,
                f"{os.path.join(enums.RunDatabases.LIVE.value, enums.RunDatabases.LIVE.value)}"
                f"{constants.DB_SEPARATOR}{live_id}",
            )
        return path

    def _merge_parts(self, *parts):
        return (
            os.path.join(*parts)
            if self.database_adaptor.is_file_system_based()
            else constants.DB_SEPARATOR.join(*parts)
        )
